@@ -22,7 +22,7 @@
 
 #include "splineimpl.h"
 
-OSL_NAMESPACE_ENTER
+OSL_NAMESPACE_BEGIN
 namespace __OSL_WIDE_PVT {
 
 OSL_USING_DATA_WIDTH(__OSL_WIDTH)
@@ -987,6 +987,6 @@ __OSL_MASKED_OP3(splineinverse, Wdf, f, Wdf)(void* wout_, const char* spline_,
 }
 
 }  // namespace __OSL_WIDE_PVT
-OSL_NAMESPACE_EXIT
+OSL_NAMESPACE_END
 
 #include "undef_opname_macros.h"
